(*  -   ,  
      ,   - 
    ,    
    ,   -  ,  
        
   .  ,   -
   ((int * int) * (string * (int * int)) list) list.*)

(* type tree = Num of int         | Var of string
             | Minus of tree      | Add of tree * tree
             | Mul of tree * tree | Frac of tree * tree
             | Pow of tree * tree | Root of tree * tree;;*)

open Edecl;;
open Ertnl;;

exception Reduction of string;;

let fw_red s = raise (Reduction s);;

let zero = [((0, 1), [])];;

(* general functions *)

let qsort f l =
 let rec qsort l =
  match l with
    l1::ls -> (qsort (List.filter (fun x -> f x l1) ls))
     @ [l1] @ (qsort (List.filter (fun x -> not (f x l1)) ls))
   |[] -> [] in qsort l;;

let power_of_mul = List.fold_left (fun x (_, y) -> x +| y) zero_rational;;

let rem_similars expr =
 let rec remsim acc expr flag =
  match expr with
    (factor1, vars1)::ls -> (
      match acc with
        ((0, 1), _) -> remsim (factor1, vars1) ls true
       |(factor, vars) when vars = vars1 -> remsim (factor +| factor1, vars) ls flag
       |_ -> acc::(remsim (factor1, vars1) ls true))
   |[] -> (
      match acc with
       |((0, 1), _) -> if flag then [] else zero
       |_ -> [acc]) in
  remsim (List.hd expr) (List.tl expr) false;;

let rec comp_vars x y =
 match (x, y) with
   (x1::xs, y1::ys) when x1 = y1 -> comp_vars xs ys
  |(x1::xs, y1::ys) when fst x1 = fst y1 -> comp_rationals (snd x1) (snd y1)
  |(x1::xs, y1::ys) -> (fst x1) < (fst y1)
  |(_, []) -> true
  |([], _) -> false;;

let comp_exprs (a, b) (c, d) =
 let (p1, p2) = (power_of_mul b, power_of_mul d) in
 if p1 = p2 then
 if p1 = zero_rational then comp_rationals a c else comp_vars b d
 else comp_rationals p1 p2;;

let sort_poly expr = let res = rem_similars (qsort comp_exprs expr) in if res = [] then zero else res;;

let is_number x =
 (List.length x = 1) && (snd (List.hd x) = []);;

(* multiplication *)

let multiply (a, b) (c, d) =
 let rec mul x y =
  match (x, y) with
    (x1::xs, y1::_ ) when fst x1 < fst y1 -> x1::(mul xs y)
   |(x1::_ , y1::ys) when fst x1 > fst y1 -> y1::(mul x ys)
   |(x1::xs, y1::ys) -> let power = snd x1 +| snd y1 in if power = zero_rational then (mul xs ys) else (fst x1, power)::(mul xs ys)
   |(x, []) -> x
   |([], y) -> y in
 (a *| c, mul b d);;

let mul_exprs a b =
 List.flatten (List.map (fun x -> List.map (fun y -> multiply x y) b) a);;

(* division *)

let divide expr (dfactor, dvars) =
 let dfactor = revert_rational dfactor in
 let dvars = List.map (fun (name, power) -> (name, negate_rational power)) dvars in
 let rec div expr =
  match expr with
    l1::ls -> (multiply l1 (dfactor, dvars))::(div ls)
   |[] -> [] in div expr;;

(* power *)

let copy_expr x n =
 let rec copy n =
  if n <= 0 then [] else x::(copy (n - 1)) in copy n;;

(* root *)

let isqrt x =
 int_of_float (sqrt (float_of_int x));;

let rec power x n =
 if n <= 0 then 1 else x * (power x (n - 1));;

let root_of_int x n =
 let rec search l r =
  if (l >= r) then
   if (power l n = x) then l else
    fw_red "root is not available." else
  let i = (l + r) / 2 in
  if power i n >= x then search l i else
  search (succ i) r in
 if n <= 1 then x else search 0 (isqrt x);;

let root_of_rational (x, y) (n, m) =
 (power (root_of_int x n) m, power (root_of_int y n) m);;

let root_of_vars vars pow =
 let rec root vars =
  match vars with
    (name, power)::ls -> (name, power /| pow)::(root ls)
   |[] -> [] in root vars;;

let reduce expr =
 let rec red t =
  match t with
    Num x -> (*if x = 0 then [] else*) [((x, 1), [])]
   |Var x -> [((1, 1), [(x, (1, 1))])]
   |Minus x -> List.map (fun (x, y) -> (negate_rational x, y)) (red x)
   |Add (a, b) -> sort_poly ((red a) @ (red b))
   |Mul (a, b) -> sort_poly (mul_exprs (red a) (red b))
   |Frac(a, b) -> let b = red b in
                  if List.length b = 1 then sort_poly (divide (red a) (List.hd b)) else fw_red "division by sum."
   |Pow (a, b) -> let (a, b) = (red a, red b) in
                  if not (is_number b) then fw_red "non-numeric power." else
                  let n = (fst (List.hd b)) in
                  if n = zero_rational then [((1, 1), [])] else
                  if (snd n = 1) && (fst n > 0) then
                    sort_poly (List.fold_left (fun x y -> mul_exprs x y) a (copy_expr a (fst n - 1))) else
                  fw_red "rational power."
   |Root(a, b) -> let (a, b) = (red a, red b) in
                  if not (is_number a) then fw_red "non-numeric power of root." else
                  if List.length b > 1 then fw_red "root of POLYnomial." else
                  let n = (fst (List.hd a)) in
                  if n = zero_rational then fw_red "zero power of root." else
                  let (x, y) = List.hd b in [(root_of_rational x n, root_of_vars y n)]
 in red expr;;